#!/usr/bin/env bash
#
# Download a given patch from the patchmanager, apply it on the specified branch,
# build/install rasdaman and run the systemtest.
#
# Usage: ./ci_test.sh <profile_file> <branch> <patch_id>
#
# Return: 0 on success, non-zero otherwise
#

PROG=$(basename $0)

readonly PROFILE="/tmp/ci_test_profile.toml"

error() {
  echo "$PROG: $@" 2>&1
  echo "$PROG: Exiting." 2>&1
  exit 1
}

log() {
  echo "$PROG: $@"
}

usage() {
  echo "$@" 2>&1
  echo
  echo "Usage: ./ci_test.sh <profile_file> [ <branch> <patch_id> ]"
  echo
  echo "Download a given patch from the patchmanager, "
  echo "apply it on the specified branch, build/install rasdaman "
  echo "and run the systemtest."
  echo "Exit with 0 on success, non-zero otherwise."
  echo
  echo "Supported OS: CentOS 7, Debian 8, Ubuntu 14.04, Ubuntu 16.04"
  exit 1
}

check_args() {
  if [ -f "profiles/test/$1" ]; then
    SRC_PROFILE="profiles/test/$1"
  elif [ -f "$1" ]; then
    SRC_PROFILE="$1"
  else
    usage "Profile file '$1' not found."
  fi
  BRANCH="$2"
  PATCH_ID="$3"

  [ -f main.py ] || error "Installer main.py not found."
}

# sets PYTHONBIN to python/python2/python3, whichever is available in that order
get_python_binary() {
  PYTHONBIN=
  for b in python3 python python2; do
    if $b --version > /dev/null 2>&1; then
      PYTHONBIN=$b; break;
    fi
  done
  [ -z "$PYTHONBIN" ] && error "Please install python."
}

check_deps() {
  get_python_binary
  wget --help > /dev/null 2>&1 || error "Please install wget."
  sudo -l > /dev/null 2>&1 || error "User has no permission to execute sudo."
}

update_profile() {
  # copy profile to a temporary location
  cp "$SRC_PROFILE" "$PROFILE" || error "Failed copying '$SRC_PROFILE' to '$PROFILE'."

  [ -n "$BRANCH" ] && sed -i "s/{branch}/$BRANCH/g" "$PROFILE"
  [ -n "$PATCH_ID" ] && sed -i "s/{patch_id}/$PATCH_ID/g" "$PROFILE"
}

fix_dpkg_locks() {
  # Ubuntu 16.04 / 18.04 runs into this error
  # E: Could not get lock /var/lib/apt/lists/lock - open (11: Resource temporarily unavailable)
  # E: Unable to lock directory /var/lib/apt/lists/
  # -> So remove this lock file to fix it
  sudo killall -9 apt apt-get > /dev/null 2>&1
  sudo rm -f /var/lib/apt/lists/lock
  sudo rm -f /var/cache/apt/archives/lock
  if [ -d /var/lib/dpkg/ ]; then
    for f in /var/lib/dpkg/lock*; do
      sudo rm -f "$f"
    done
    sudo dpkg --configure -a > /dev/null 2>&1
  fi
}

run_installer() {
  sudo "$PYTHONBIN" ./main.py "$PROFILE"
}

check_args "$@"
check_deps
update_profile
fix_dpkg_locks
run_installer
