#!/usr/bin/env bash
#
# Help building nightly/stable/testing packages.
#

PROG=$(basename $0)

readonly SUPPORTED_OPTS="(nightly|stable|testing)"
readonly CENTOS7="centos7"
readonly CENTOS8="centos8"
readonly CENTOS7GDAL2="centos7_gdal2"
readonly JESSIE="jessie"
readonly BUSTER="buster"
readonly STRETCH="stretch"
readonly TRUSTY="trusty"
readonly XENIAL="xenial"
readonly BIONIC="bionic"
readonly FOCAL="focal"
readonly JAMMY="jammy"

# global variables
pkg_type=
repo_type=
os=

error() { echo "$PROG: $@" 2>&1; exit 1; }
log()   { echo "$PROG: $@"; }
usage() {
  echo "$@" 2>&1
  echo
  echo "Usage: ./pkg_build.sh $SUPPORTED_OPTS [<branch>] [noupload]"
  echo
  echo "Build nightly/stable/testing packages from specified branch."
  echo "Supported OS: CentOS 7, Debian 8/9, Ubuntu 14.04/16.04/18.04"
  exit 1
}

check_root() {
  [ "$EUID" -eq 0 ] || error "This script has to be executed with root user."
}
# sets PYTHONBIN to python/python2/python3, whichever is available in that order
get_python_binary() {
  PYTHONBIN=
  for b in python python2 python3; do
    if $b --version > /dev/null 2>&1; then
      PYTHONBIN=$b; break;
    fi
  done
  [ -z "$PYTHONBIN" ] && error "Please install python."
}
check_deps() {
  wget --help > /dev/null 2>&1 || error "Please install wget."
  get_python_binary
}

get_os() {
  log "Checking OS..."
  if [ -f "/etc/centos-release" ]; then
    grep -q "CentOS Linux release 7" /etc/centos-release
    if [ $? -eq 0 ]; then
      if [ -f "/usr/lib/libgdal.so.20" ]; then
        os=$CENTOS7GDAL2
      else
        os=$CENTOS7
      fi
    else
      grep -q "CentOS Linux release 8" /etc/centos-release
      if [ $? -eq 0 ]; then
        os=$CENTOS8
      else
        error "$(cat /etc/centos-release) is not supported."
      fi
    fi
    repo_type="rpm"
  else
    tomcat="tomcat7"
    local version=$(lsb_release -a 2>&1 | grep Description \
      | sed 's/Description: *//' | tr -d '[:space:]' || error "Unsupported OS.")
    case "$version" in
      Ubuntu14.*) os=$TRUSTY;;
      Ubuntu16.*) os=$XENIAL; tomcat="tomcat8";;
      Ubuntu18.*) os=$BIONIC; tomcat="tomcat9";;
      Ubuntu20.*) os=$FOCAL; tomcat="tomcat9";;
      UbuntuJammy*) os=$JAMMY; tomcat="tomcat9";;
      Debian*8*)  os=$JESSIE;;
      Debian*9*)  os=$STRETCH; tomcat="tomcat8";;
      Debian*buster*)  os=$BUSTER; tomcat="tomcat9";;
      *) error "Unsupported OS $version";;
    esac
    repo_type="deb"
  fi
  log "Done, detected OS '$os'."
}

clear_apt_locks() {
  # Ubuntu 16.04/18.04 runs into this error
  # E: Could not get lock /var/lib/apt/lists/lock - open (11: Resource temporarily unavailable)
  # E: Unable to lock directory /var/lib/apt/lists/
  # -> So remove this lock file to fix it
  sudo killall -9 apt apt-get > /dev/null 2>&1
  sudo rm -f /var/lib/apt/lists/lock
  sudo rm -f /var/cache/apt/archives/lock
  for f in /var/lib/dpkg/lock*; do
    sudo rm -f "$f"
  done
  sudo dpkg --configure -a
}
build_package() {
  [ "$repo_type" == "deb" ] && clear_apt_locks

  local PROFILE_TEMPLATE="profiles/package/generate.toml"
  local PROFILE="profiles/package/generate.toml.instantiated"
  typeset -i ITERATION="`wget --no-check-certificate -qO- https://download.rasdaman.org/packages/iteration_$pkg_type.txt`"
  sed "s/{iteration}/${ITERATION}/g" "${PROFILE_TEMPLATE}" > "${PROFILE}"
  sed -i "s/version = \"master\"/version = \"${branch}\"/g" "${PROFILE}"

  echo "Building $pkg_type package from branch $branch, iteration $ITERATION..."
  "$PYTHONBIN" main.py "$PROFILE"
}

upload_package() {
  [ "$upload" = true ] || return
  # Copy package in the download area
  local non_root_user=$(grep 1000 /etc/passwd | cut -d: -f1)
  target_dir="/mnt/rasservice/Download/package_source/$repo_type/$os/$pkg_type"
  log "Uploading package with user $non_root_user to directory download@download.rasdaman.org:$target_dir..."
  log " package: "
  ls /tmp/rasdaman-built-packages

  local private_key=/home/$non_root_user/.ssh/id_rsa
  chmod 600 $private_key
  scp -i $private_key -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -r \
      /tmp/rasdaman-built-packages/* download@download.rasdaman.org:"$target_dir"
  if [ $? -ne 0 ];then
      echo "ERROR: Failed to remote copy!"
      exit 1
  fi
}

# ------------------------------------------------------------------------

# parse arguments
#
# $1: package type (stable/nightly)
pkg_type="$1"
shift
[ -n "$pkg_type" ] || usage "No package type (nightly/stable/testing) specified."
echo "$SUPPORTED_OPTS" | grep "$pkg_type" -q || usage "Invalid package type."
# $2: branch, if specified
[ $# -ge 1 ] && { branch="$1"; shift; } || branch="master"
# $3: whether to upload the package or not
[ $# -ge 1 -a "$1" == "noupload" ] && upload=false || upload=true

check_root
check_deps
get_os
build_package && upload_package

log "Done."
