"""
 *
 * This file is part of rasdaman community.
 *
 * Rasdaman community is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Rasdaman community is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU  General Public License for more details.
 *
 * You should have received a copy of the GNU  General Public License
 * along with rasdaman community.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2003 - 2016 Peter Baumann / rasdaman GmbH.
 *
 * For more information please see <http://www.rasdaman.org>
 * or contact Peter Baumann via <baumann@rasdaman.com>.
 *
"""
import os
from abc import abstractmethod
from util.log import log
from util.commands import make_directory, remove
from util.string_util import strip_whitespace
from services import executor


class Retriever:
    def __init__(self):
        """
        A retriever is in charge of providing the source code of rasdaman in
        a specific state, e.g. from the repository, from a patch etc
        """
        pass

    @abstractmethod
    def retrieve(self):
        """
        Retrieves the source code using its internal mechanism
        """
        pass


class RetrieverDecorator(Retriever):
    def __init__(self, retriever):
        """
        Decorator for retrievers, helps the composition process
        :param Retriever retriever: the retriever to decorate
        """
        Retriever.__init__(self)
        self.retriever = retriever

    def retrieve(self):
        self.retriever.retrieve()


class DoNothingRetriever(Retriever):
    def __init__(self):
        """
        Useful to keep strategies sane when no retrieving is done
        """
        Retriever.__init__(self)

    def retrieve(self):
        Retriever.retrieve(self)


class RepositoryRetriever(Retriever):
    def __init__(self, repo_url, location, version="master", shallow=False):
        """
        The retriever is in charge of providing the source code of rasdaman
        from the main repository at the latest commit on the specified branch
        :param str repo_url: the repository url of rasdaman
        :param str location: path where the source should be downloaded
        :param str version: the branch/tag to be retrieved
        """
        Retriever.__init__(self)
        self.repo_url = repo_url
        self.location = location
        self.version = version
        self.shallow = shallow

    def retrieve(self):
        """
        Retrieve the rasdaman source code
        :return: the path where rasdaman source code was saved
        :rtype: str
        """
        # Make sure the directory in which we retrieve exists and create it if not
        if not os.path.exists(self.location):
            make_directory(self.location)

        # Clone the rasdaman repository there
        if os.path.exists(self.repo_url):
            # the source repo is a local directory, rsync it to the source location
            log.info("Syncing local rasdaman repository at {} to {}"
                     .format(self.repo_url, self.location))
            src = self.repo_url
            if not src.endswith('/'):
                src = src + '/'
            cmd = ["rsync", "-a", src, self.location]
            executor.execute(cmd,  self.location, timeout_sec=300)
        elif len(os.listdir(self.location)) == 0:
            log.info("Cloning rasdaman repository at {0}.".format(self.location))
            cmd = ["git", "clone", "--quiet", "--branch", self.version]
            if self.shallow:
                cmd = cmd + ["--depth", "400"]
            executor.execute(cmd + [self.repo_url, "."],  self.location, timeout_sec=600)
        else:
            if not os.path.exists(self.location + "/.git"):
                log.warn("No git repository found at '" + self.location +
                         "', skipping the rasdaman source tree update.")
                return

            log.info("Updating rasdaman source code at {0}".format(self.location))
            cmd = ["git", "pull", "--quiet"]
            if self.shallow:
                cmd = cmd + ["--depth=400"]
            executor.execute(cmd, self.location,
                             throw_on_error=False, warn_on_error=True, timeout_sec=600)
            # Checkout the selected branch
            executor.execute(["git", "checkout", "--quiet", self.version], self.location,
                             throw_on_error=False, warn_on_error=True)

        return self.location


class PatchManagerRetriever(RetrieverDecorator):
    def __init__(self, retriever, source_path, patch_url):
        """
        Retrieves a patch from a specified url and applies to the rasdaman source code
        :param Retriever retriever: the retriever to decorate
        :param str source_path: the path to the source
        :param str patch_url: the url to the patch
        :return:
        """
        RetrieverDecorator.__init__(self, retriever)
        self.source_path = source_path
        self.patch_url = patch_url

    def __set_git_user_config(self):
        """
        user.name and user.email must be set for git am to work
        """
        _, _, rc = executor.execute(["git", "config", "user.name"],
                                    self.source_path, throw_on_error=False)
        if rc != 0:
            executor.execute(["git", "config", "user.name", "'Rasdaman Installer'"],
                             self.source_path, throw_on_error=False)
            executor.execute(["git", "config", "user.email",
                             "RasdamanInstaller@rasdaman.com"],
                             self.source_path, throw_on_error=False)

    def __download_patch(self):
        if self.patch_url[0] == '/':
            return self.patch_url
        patch_archive_path = "{}/patch.tgz".format(self.source_path)
        executor.execute(["wget", "--no-check-certificate", "--quiet", self.patch_url, "-O", patch_archive_path],
                         self.source_path)

        patch_path, _, _ = executor.execute(["tar", "-xzvf", patch_archive_path],
                                            self.source_path)
        remove(patch_archive_path)
        return strip_whitespace(patch_path)

    def __apply_patch(self, patch_path):
        self.__set_git_user_config()
        # let it fail if patch is not there or does not apply correctly
        executor.execute(["git", "am", "-3", "--ignore-whitespace", patch_path],
                         self.source_path)
        if patch_path[0] != '/':
            remove(patch_path)

    def retrieve(self):
        RetrieverDecorator.retrieve(self)
        log.info("Retrieving and applying patch from " + self.patch_url)
        patch_path = self.__download_patch()
        self.__apply_patch(patch_path)
        log.info("Patch retrieved and applied successfully.")
